#include "BF_GJ.h"
//#include <../config.h>
#include <stdio.h>
#include <iostream>

namespace BF
{
    GJ::GJ(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed, bool _muAdp, double _mu, UtUpdateType _separationMatUpdateType) : 
    BF(_numFreq, _lowerfreq, _upperfreq, _samplingRate, _soundSpeed, _separationMatUpdateType), 
    muAdp(_muAdp), 
    mu(_mu)
    {
    }

    GJ::~GJ()
    {
    }

    void
    GJ::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
/*== source from adaptBF.m

  % [2] L. Griffiths and C. Jim, 'An alternative approach to linearly
  % constrained adaptive beamforming', IEEE trans. on ant. and propag.
  % Vol. AP-30, No.1, 1982
  % This uses TF null vectors for blocking matrix (Ws in Eq. 26 in [2])
  % The difference between this program and the paper are: 
  % 1: This uses complex. Transpose is processed with complex conjugation.
  % 2: This uses row vectors for BF coefficient. (Paper uses column vectors)
  % 3: Tapped delay pre-process in the first stage is done in Wc and Ws 
      if isempty(Adata.Ws) % Initialization process
        Adata.Wc = zeros(NSrc, NMic, NFreq); % Main fixed BF coefficients (Wc)
        Adata.Ws = zeros(NMic-1, NMic, NSrc, NFreq); % Blocking matrix (Ws)
        Adata.Al = zeros(NSrc, NMic-1, NFreq); % Adaptive LMS coefficients (Al)
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
            h = Adata.H(:, ns, nf); % TF vector
            if isempty(G) % If initial G is not defined
              Adata.Wc(ns, :, nf) = pinv(h);
            else % If initial G is defined
              Adata.Wc(ns, :, nf) = G(ns, :, nf); % Set G as fixed-BF part
            end
              [U, S, V] = svd(h);
              Adata.Ws(:, :, ns, nf) = (U(:, S==0))';
          end
        end
        if isempty(G), G = Adata.Wc; end
        return
      else % update process
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
              Wc = Adata.Wc(ns, :, nf);
              Ws = Adata.Ws(:, :, ns, nf);
              Al = Adata.Al(ns, :, nf);
              x = Adata.x(:, nf);
              Xp = Ws*x; % Filtered input by blocking matrix
              y = Adata.y(ns, nf); % Total output
              myu = Adata.myu;
              Al = Al + myu*y*Xp'; % Al update by LMS (Eq. 30 in [2])
              g = Wc - Al*Ws; % Total BF coef. based on Fig.4
              Adata.Al(ns, :, nf) = Al; % Store adaptive coef.
              G(ns, :, nf) = g; % Store total BF coef.
          end
        end

        return
      end % For if initialization
      
*/
	int numSrc = sourcePositions.size();

	/*== 直近フレーム・分離行列データベースの音源と、このフレームでの音源との対応をとる ==*/
	std::vector<int> sourceIndices;
	std::vector<int> sourceIndicestypes;
	getSourceIndex(sourcePositions, sourceIndicestypes, sourceIndices);	

	/*== 伝達関数を再計算する ==*/
	std::vector<Matrix> newHs, newGs;
	this->BF::getTransformFunctions(sourcePositions, newHs);

	/*== 分離行列の引き継ぎまたは初期値の計算 ==*/
	std::vector<Matrix> newWcs, newWss, newAls;
//	for(int i = lowerfreq; i <= upperfreq ; i++){
	for(int i = 0; i < this->numFreq; i++){
		const Matrix &H = newHs[i];
	    int numMic = H.getRowCount();

		Matrix newWc(numSrc, numMic);
		Matrix newAl(numSrc, numMic - 1);
		Matrix newG (numSrc, numMic);
	    
		// 対象周波数のみ分析
	    if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
			newWc = Matrix::O(numSrc, numMic);
			newAl = Matrix::O(numSrc, numMic - 1);
			newG = Matrix::O(numSrc, numMic);
		    for(int j = 0; j < numSrc; j++) {
				Matrix newWs = Matrix::O(numMic - 1, numMic);
				newWss.push_back(newWs);
			}
		}
		else {
		    for(int j = 0; j < numSrc; j++){
				Matrix h = H.getColumn(j);	// Hは通常の伝達関数となっている

				const SourceLocation &srcPos = sourcePositions[j];

				/* 入力の j 番目の音源と同じ ID の音源が、直近フレームにあるか、データベースにあるか、見当たらないか？ */
				const int sourceIndex = sourceIndices[j];
				const int type = sourceIndicestypes[j];
				/*if (type == 0){
					// 前フレームのデータからの引き継ぎ
					newG.setRow(j, this->separationMatrices[i].getRow(sourceIndex));
					newWc.setRow(j, this->Wc[i].getRow(sourceIndex));
				}
				else if (type == 1){
					// データベースからの引き継ぎ
  					setSeparationMatricesFromDatabase(sourceIndex, j, i, newG);
					newWc.setRow(j, h.pseudoInverse()); // Wcは初期値, h は列ベクトルなので、pinv(h) = h' / |h|^2 になる。
				}
				else {
					// 初期値を入力
					newG.setRow(j, h.pseudoInverse()); // h は列ベクトルなので、pinv(h) = h' / |h|^2 になる。
					newWc.setRow(j, h.pseudoInverse()); // Wcは初期値, h は列ベクトルなので、pinv(h) = h' / |h|^2 になる。
				}*/
				if (type == 0) { // 直近フレームにある
					// 直近フレームのデータからの引き継ぎ
					newWc.setRow(j, this->Wc[i].getRow(sourceIndex));
					newG.setRow(j, this->separationMatrices[i].getRow(sourceIndex));
				}
				else { // データベースにある, 見当たらない
					Matrix pinvh = h.pseudoInverse(); // h は列ベクトルなので、pinv(h) = h' / |h|^2 になる

					// Wcは初期値
					newWc.setRow(j, pinvh);

					if (type == 1) { // データベースにある
						// データベースからの引き継ぎ
	  					setSeparationMatricesFromDatabase(sourceIndex, j, i, newG);
					}
					else { // (type == -1) 見当たらない
						// 初期値を入力
						newG.setRow(j, pinvh);
					}
				}
		
				/*
				 * h (列ベクトル) を特異値分解。
				 *
				 * この場合、
				 *   u = [-h/|h|, *, *]
				 *   s = [|h|, 0, 0]
				 *   v = [-1]
				 * になる。
				 */
				Matrix u, s, v;
				/*
				 * 特異値分解の結果、u は octave 版とは異なる値になる。
				 * 具体的には、s が 0 の要素に対応する列について、値が異なる。
				 * s=0 となる要素は複数あるため、これらに対応する u は一意ではないかも。
				 * Ws の要素として、これらを用いることから、後の値は octave 版と一致しない。
				 */
				h.singularValueDecomposition(u, s, v);
				//h.dump(stdout, "h");
				//u.dump(stdout, "u");
				//s.dump(stdout, "s");
				//v.dump(stdout, "v");

				Matrix newWs(numMic - 1, numMic);
				/* h は rank=1 なので、S は (+, 0, 0) の形
				   -> U(:, S==0) -> U の 1 列目以外を採用。 */
				for(int k = 0; k < newWs.getRowCount(); k++){
					newWs.setRow(k, u.getColumn(k + 1).transposed());
				}
				//h.pseudoInverse().dump(stdout, "pinv(h)");
				//newG.dump(stdout, "newG");
				
				newWss.push_back(newWs);
			}
		}

		newGs. push_back(newG);
		newAls.push_back(newAl);
		newWcs.push_back(newWc);
	}

	/*== 分離行列データベースの座標を更新 ==*/
	updateSeparationMatricesDatabaseSource(sourcePositions, sourceIndicestypes, sourceIndices);
#if 0
int ifreqindex = 1;
std::cout << "size of newGS[" << ifreqindex << "] = (" << newGs[ifreqindex].getRowCount() << "," << newGs[ifreqindex].getColumnCount() << ")";
for (int k = 0 ; k < sourcePositions.size() ; k++){
std::cout << "BF->G(f" << ifreqindex << "s" << sourcePositions[k].id << "):";
for (int j = 0 ; j < this->micPositions.size() ; j++){
std::cout << newGs[ifreqindex].getElementAt(k, j) << ",";
}
std::cout << "@BF_GJ::updateSources" << std::endl;
}
#endif

	/*== 結果の格納 ==*/
	this->sourcePositions = sourcePositions;
	newWcs.swap(this->Wc);
	newWss.swap(this->Ws);
	newAls.swap(this->Al);
	newHs.swap(this->transformFunctions);
	newGs.swap(this->separationMatrices);
    }

    void
    GJ::adaptSeparationMatrices()
    {
	int numSrc = this->sourcePositions.size();

	std::vector<Matrix> newGs;
	std::vector<Matrix> newAls;
//	for(int i = lowerfreq; i <= upperfreq ; i++){
	for(int i = 0; i < this->numFreq; i++){
	    const Matrix &H = this->transformFunctions[i];
	    int numMic = H.getRowCount();

		// 対象周波数のみ分析
//    	if (i < lowerfreq || i > upperfreq){
	    if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
	        newAls.push_back(Matrix::O(numSrc, numMic - 1));
	        newGs.push_back(Matrix::O(numSrc, numMic));
			continue;
	    }

	    Matrix newG (numSrc, numMic);
	    Matrix newAl(numSrc, numMic - 1);
	    for(int j = 0; j < numSrc; j++){
			const Matrix  Wc = this->Wc[i].getRow(j);
			const Matrix &Ws = this->Ws[i * numSrc + j];
			const Matrix  x  = inputSpectrum.getColumn(i);
			const Matrix  Xp = Ws * x;
			const Complex y  = this->outputSpectrum.getElementAt(j, i);
			// const Complex mu = this->mu;

			Complex mu;
			if(this->muAdp){
			  // mu = norm(y) / (2.0 * (norm(y) * Xp.norm())) * this->mu; // original adaptive stepsize
				if(abs(Xp.norm().real()) < BF_EPS){
					mu = 1.0 / (2.0 * Xp.norm().real() + BF_EPS ) * this->mu; // norm(y) is cancelled
				}else{
					mu = 1.0 / (2.0 * Xp.norm().real()) * this->mu; // norm(y) is cancelled
				}
			}else{
			  mu = this->mu;
			}

#if ENABLE_JDEBUG
			double Jc = 1.0 / 2.0 * norm(y);
			if(j==0) std::cout << mu.real() << " " << Jc << " ";		
#endif

			const Matrix  AlRow = this->Al[i].getRow(j) + mu * y * Xp.transposed();
			const Matrix  g  = Wc - AlRow * Ws;

			//Wc.dump(stdout, "Wc");
			//Ws.dump(stdout, "Ws");
			//AlRow.dump(stdout, "AlRow");

			newAl.setRow(j, AlRow);
			newG.setRow(j, g);
	    }
	    newAls.push_back(newAl);
	    newGs.push_back(newG);
	}

#if ENABLE_JDEBUG
	std::cout << std::endl;
#endif

	/*== 結果の格納 ==*/
	newAls.swap(this->Al);
	newGs.swap(this->separationMatrices);
    }
}
