#include "BF_MVDR.h"
#include <stdio.h>
#include <iostream>

namespace BF
{
	CMVDR::CMVDR(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed, double _regFactor) :
		BF(_numFreq, _lowerfreq, _upperfreq, _samplingRate, _soundSpeed, UtID), 
		regFactor(_regFactor)
	{
	}

	CMVDR::~CMVDR()
	{
	}

	void CMVDR::updateSources(const std::vector<SourceLocation> &sourcePositions)
	{
/*== source from CHiME3_enhance_data.m
        % STFT
        X = stft_multi(x.',wlen);
        [nbin,nfram,~] = size(X);
        
        % Compute noise covariance matrix
        N=stft_multi(n.',wlen);
        Ncov=zeros(nchan,nchan,nbin);
        for f=1:nbin,
            for n=1:size(N,2),
                Ntf=permute(N(f,n,:),[3 1 2]);
                Ncov(:,:,f)=Ncov(:,:,f)+Ntf*Ntf';
            end
            Ncov(:,:,f)=Ncov(:,:,f)/size(N,2);
        end
        
        % Localize and track the speaker
        [~,TDOA]=localize(X);
        
        % MVDR beamforming
        Xspec=permute(mean(abs(X).^2,2),[3 1 2]);
        Y=zeros(nbin,nfram);
        for f=1:nbin,
            for t=1:nfram,
                Xtf=permute(X(f,t,:),[3 1 2]);
                Df=sqrt(1/nchan)*exp(-2*1i*pi*(f-1)/wlen*fs*TDOA(:,t)); % steering vector
                Y(f,t)=Df(~fail)'/(Ncov(~fail,~fail,f)+regul*diag(Xspec(~fail,f)))*Xtf(~fail)/(Df(~fail)'/(Ncov(~fail,~fail,f)+regul*diag(Xspec(~fail,f)))*Df(~fail));
            end
        end
        y=istft_multi(Y,nsampl).';
*/

		const int numSrc = sourcePositions.size();

		/*== 伝達関数を再計算する ==*/
		std::vector<Matrix> newHs;
		this->BF::getTransformFunctions(sourcePositions, newHs);
		
		/*== 分離行列を計算する ==*/
		std::vector<Matrix> newGs;
		for (int i = 0; i < this->numFreq; i++) {
			/* 伝達関数 */
			const Matrix &A = newHs[i]; // Row:numMic, Column:numSrc
			int numMic = A.getRowCount();

			/* 対象周波数のみ計算 */
	    	if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
				newGs.push_back(Matrix::O(numSrc, numMic));
				continue;
			}

			/* 相関行列 */
			Matrix K = this->covarianceMatrices[i];
			K = K + this->regFactor * this->inputPowerSpectrum.getColumn(i).diagonalMatrix();
			Matrix Kinv = K.pseudoInverse(); /*= K.Inverse();*/ // Kの逆行列

			/* MVDR Beamformer */
			Matrix G(numSrc, numMic);
			for (int j = 0; j < numSrc; j++) {
				/* (K(-1) * Tf) / (Tf' * K(-1) * Tf) */

				Matrix Tf = A.getColumn(j) / sqrt(numMic);
				Matrix Tft = Tf.transposed(); // Tfの複素共役転置

				Matrix TftKinv = Tft * Kinv;
				Matrix TftKinvTf = TftKinv * Tf;

				Complex sclrTftKinvTf = TftKinvTf.scalar();	// 1x1 の行列をスカラー値に変換

				Matrix W; // W [rows:numMic, cols:1]
				if (sclrTftKinvTf == 0.0)
					W = TftKinv * 0.0;
				else
					W = TftKinv / sclrTftKinvTf;
				
				G.setRow(j, W);
			}
			newGs.push_back(G);
		}

		/*== 結果の格納 ==*/
		this->sourcePositions = sourcePositions;
		newHs.swap(this->transformFunctions);
		newGs.swap(this->separationMatrices);
	}

	void CMVDR::adaptSeparationMatrices()
	{
		/* do nothing */
	}
}
